const _ = require('lodash');
const async = require('async');
const DEBUG = require('debug');

const Error = require('./error');
const config = require('./config');
const handlers = require('./request');
const interfaces = require('./interfaces');

class FacePlatform {
    constructor (option) {
        option = option || {};
        if (!config.versions.includes(option.version)) {
            throw Error(80001);
        }
        option.address = option.address || config.default_address;
        this.address = option.address.startsWith('http') ? option.address : 'http://' + option.address;
        this.version = option.version;
        this.auth = option.auth;
        this.request_timeout = option.request_timeout || config.request_timeout;
        this.chains = [];
        this.refreshing = false;        // is refreshing session_id
        this.revoking = false;          // is the chaining executing
        this.revoked = false;           // have any interfaces been revoked in the past time cycle
        this.debug = DEBUG("faceplatform");

        const fp = new Proxy(this, {
            get: (target, key) => {
                if (interfaces[key]) {
                    return (...args) => {
                        this.chains.push(handlers.requestHandler.bind(target, key, ...args));
                        return fp;
                    };
                } else {
                    return target[key];
                }
            }
        });

        this.refreshSession();
        return fp;
    }

    exec () {
        this.revoking = true;
        this.debug('FacePlatform address: ' + this.address + ' with version: ' + this.version);
        this.debug('start revoking');

        let p = Promise.resolve();
        while (this.chains.length) {
            const handler = this.chains.shift();
            const wrap_f = () => {
                if (this.to_break) {
                    return Promise.reject(Error(80201));
                }

                const f = (resolve) => {
                    this.revoked = true;
                    let result = handler();
                    if (result instanceof FacePlatform) {
                        // somehow the chains auto breaks if an instance of FP is returned
                        this.debug('this chain function returns an instance of FacePlatform, which is not allowed, undefined is returned instead.');
                        result = undefined;
                    }
                    resolve(result);
                };

                // wait while it's refreshing session_id
                return new Promise((resolve, reject) => {
                    async.whilst(() => {
                        return this.refreshing;
                    }, (next) => {
                        this.debug('fp is refreshing session_id, wait for 100ms.');
                        setTimeout(next, 100);
                    }, () => {
                        f(resolve);
                    });
                });
            };
            p = p.then(wrap_f);
        }
        p.catch(err => {
            if (config.forgivable_errors.includes(err.error_code)) {
                this.debug('forgivable error: %s', err.message);
            } else {
                this.throw(err);
            }
        })
        .then(() => {
            this._done();
        });
    }

    _done () {
        this.revoking = false;
        this.to_break = false;
        this.chains = [];
        const _f = this._finally;
        delete this._finally;
        this.debug('_done revoking');
        _f();
    }

    break () {
        this.debug('manually stop chaining here');
        this.to_break = true;
    }

    then (next) {
        this.chains.push(next.bind(this));
        return this;
    }

    throw (err) {
        this.debug(err);
        if (_.isFunction(this._catch)) {
            this._catch(err);
            delete this._catch;
        } else {
            throw err;
        }
    }

    catch (next) {
        this._catch = next;
        return this;
    }

    finally (next) {
        this._finally = next;
        return this;
    }

    w (func) {
        const _w_f = () => {
            return func();
        }
        _w_f.w = true;
        return _w_f;
    }

    refreshSession () {
        const finish = () => {
            this.revoked = false;
            this.refreshing = false;
            this.debug("end refreshing session.");
        };

        const doRefresh = () => {
            this.debug("start refreshing session.");
            if (this.session_id && this.revoked) {
                this.debug("something has been revoked in the past cycle, so skip it.");
                finish();
            } else {
                this.refreshing = true;
                handlers.login.apply(this)
                    .then(finish);
            }
        };

        doRefresh();
        this.refreshInterval = setInterval(doRefresh, config.refresh_interval);
    }
};

module.exports = FacePlatform;
